﻿/// @file mp_printer.cpp
///
/// @brief MPPRINTER 制御クラス
///
/// @date 2013.11.27 Create
///
/// @author Copyright (c) 2013, Sasaji. All rights reserved.
///
#include "mp_printer.h"
#include "mp1020.h"
#include "mp80.h"
#include <wx/filename.h>
#include <wx/wfstream.h>

#ifdef __WXOSX__
#include "wx/osx/printdlg.h"
// 描画範囲外のデータをトリミング
//#define USE_TRIMING_AREA
#define PRINT_FRAMEWORK 1
#else
#define PRINT_FRAMEWORK 2
#endif
#ifdef _DEBUG_LOG
static const wxString capsTypeStr[4] = {_T("Display"),_T("Printer"),_T("Preview"),_T("?")};
#endif

#ifndef USE_PIXELDATA
#define MX_RGB_SIZE	3
#endif

MP_PRINTER::MP_PRINTER(const wxApp &app, const wxString &data_path)
{
	mApp     = &app;
	mDataPath = data_path;

	memset(&rDw, 0, sizeof(rDw));
	memset(&rDwSz, 0, sizeof(rDwSz));
	bufDw = NULL;

	memset(&rMx, 0, sizeof(rMx));
	memset(&rMxMax, 0, sizeof(rMxMax));
	memset(&rMxMaxV, 0, sizeof(rMxMaxV));
#ifdef USE_PIXELDATA
	bmpMx = NULL;
#endif
	bufMx = NULL;
	dcMx = NULL;

#ifdef USE_CASH_BUFFER
	mUseCash = false;
	mDataCashd = false;
	bmpCashd = NULL;
	dcCashd = NULL;
#endif
	zoom_mag = 100;


	mPrintDC = NULL;

	mp_model = NULL;
	mp_model_list[MP_MODEL_MP1020] = new MP1020(this);
	mp_model_list[MP_MODEL_MP80] = new MP80(this);

	for(int i=0; i<capsUnknownType; i++) {
		device_info[i].tech = (CapsInfoTypeTech)i;
	}

#if defined(__WXOSX__)
	// MacではANDがエラーになる
	device_info[capsDisplayType].mBlitMode = wxCOPY;
	device_info[capsPrinterType].mBlitMode = wxCOPY;
	device_info[capsPreviewType].mBlitMode = wxCOPY;
#elif defined(__WXGTK__)
	device_info[capsDisplayType].mBlitMode = wxAND;
	device_info[capsPrinterType].mBlitMode = wxCOPY;
	device_info[capsPreviewType].mBlitMode = wxCOPY;
#else
	device_info[capsDisplayType].mBlitMode = wxAND;
	device_info[capsPrinterType].mBlitMode = wxCOPY;
	device_info[capsPreviewType].mBlitMode = wxCOPY;
#endif
}

MP_PRINTER::~MP_PRINTER()
{
	release_cashdata();
#ifdef USE_PAGE_BUFFER
	delete dcPage;
#endif
	delete dcMx;
#ifdef USE_PIXELDATA
	delete bufMx;
	delete bmpMx;
#else
	delete[] bufMx;
#endif
	delete[] bufDw;
	delete mp_model_list[MP_MODEL_MP80];
	delete mp_model_list[MP_MODEL_MP1020];
}

// 初期化
bool MP_PRINTER::Init(wxWindow *win, int w, int h)
{
	int i;

	bufDw = new DpType[MP_PRN_BUF_WIDTH * MP_PRN_BUF_HEIGHT];
#ifdef USE_PIXELDATA
	bmpMx = new wxBitmap(MP_PRN_BUF_WIDTH, MP_PRN_BUF_HEIGHT, 24);
	if (!bmpMx->IsOk()) {
		gErrInfo.SetInfo(__FILE__, __LINE__, myError, myErrSystem, _T("Cannot create bmpMx."));
		gErrInfo.ShowMsgBox(win);
		return false;
	}
	bufMx = new wxNativePixelData(*bmpMx);
	if (bufMx->GetWidth() != MP_PRN_BUF_WIDTH
	 || bufMx->GetHeight() != MP_PRN_BUF_HEIGHT) {
		gErrInfo.SetInfo(__FILE__, __LINE__, myError, myErrSystem, _T("Invalid create bufMx."));
		gErrInfo.ShowMsgBox(win);
		return false;
	}
#else
	bufMx = new wxByte[MP_PRN_BUF_WIDTH * MP_PRN_BUF_HEIGHT * MX_RGB_SIZE];
#endif
	wxBitmap bmpDummy(MP_PRN_BUF_WIDTH, MP_PRN_BUF_HEIGHT, 24);
	if (!bmpDummy.IsOk()) {
		gErrInfo.SetInfo(__FILE__, __LINE__, myError, myErrSystem, _T("Cannot create bmpDummy."));
		gErrInfo.ShowMsgBox(win);
		return false;
	}
	dcMx = new wxMemoryDC(bmpDummy);
	if (!dcMx->IsOk()) {
		gErrInfo.SetInfo(__FILE__, __LINE__, myError, myErrSystem, _T("Cannot create dcMx."));
		gErrInfo.ShowMsgBox(win);
		return false;
	}
//	dcMx->SelectObjectAsSource(*bmpMx);
	dcMx->SetBackground(wxBrush(wxColour(0xff,0,0)));
	dcMx->Clear();

#ifdef USE_PAGE_BUFFER
	dcPage = NULL;
#endif

	SetWindowSize(w, h);
	ClearData(_T(""));

	for (i=MP_MODEL_MP1020; i<=MP_MODEL_MP80; i++) {
		// read font data
		if (mp_model_list[i]->ReadFont() != true) {
			gErrInfo.SetInfo(__FILE__, __LINE__, myError, myErrFontFile);
			gErrInfo.ShowMsgBox(win);
			return false;
		}
		mp_model_list[i]->ClearSetting();
		mp_model_list[i]->PushSetting();
	}
	SelectModel(MP_MODEL_MP1020);
	ClearSetting();

#ifdef USE_CASH_BUFFER
	mUseCash = gConfig.GetMemoryCash();
#endif

	return true;
}

// 機種選択
void MP_PRINTER::SelectModel(int model)
{
	mp_model = mp_model_list[model];
	selected_model = model;

	for(int i=0; i<capsUnknownType; i++) {
		device_info[i].vm_ppi.x = mp_model->GetPPIX();
		device_info[i].vm_ppi.y = mp_model->GetPPIY();
	}
	rDpPpi.x = mp_model->GetPPIX();
	rDpPpi.y = mp_model->GetPPIY();

	mp_model->GetDispAspectRatio(&dsr);

	Zoom(NULL, zoom_mag);
}

// 紙サイズなどを初期状態に戻す
void MP_PRINTER::ClearSetting()
{
	mp_model->ClearSetting();

	// A4
	printdata.SetOrientation(wxPORTRAIT);
	printdata.SetPaperId(wxPAPER_A4);
	printdata.SetNoCopies(1);

	printdlg.SetPrintData(printdata);
	pagesetdlg.SetPrintData(printdata);

	// 紙の余白
	pagesetdlg.SetMarginTopLeft(wxPoint(10,10));	// 10mm
	pagesetdlg.SetMarginBottomRight(wxPoint(10,10));	// 10mm
	for(int i=0; i<capsUnknownType; i++) {
		device_info[i].margin_mm.top = 10;
		device_info[i].margin_mm.bottom = 10;
	}

	pagesetdlg.SetMinMarginTopLeft(wxPoint(0,0));	// 0mm
	pagesetdlg.SetMinMarginBottomRight(wxPoint(0,0));	// 0mm

	return;
}

// データバッファクリア
void MP_PRINTER::ClearData(const wxString &title)
{
	text_file = title;
	memset(text_data, 0, sizeof(text_data));
	text_len = 0;

	draw_mode = false;
	draw_mode_bkup = false;
	printing = 0;
}

// データバッファにセット
void MP_PRINTER::SetData(wxWindow *win, wxByte* data, int size)
{
	int first_len = text_len;

	for(int pos = 0; pos < size && text_len < sizeof(text_data); pos++) {
		text_data[text_len] = data[pos];
		// データは0,1が反転している
		text_data[text_len] = ~text_data[text_len];
		text_len++;
	}

	// 画面に反映
	if (first_len == 0) {
		mp_model->PushSetting();
		draw_mode = true;
		if (win != NULL) Preview(win);
	}
}

//
// ページ設定ダイアログ
//
void MP_PRINTER::PageSetup(wxWindow *win)
{
	/* ページ設定ダイアログ */
	if (set_printer_page(win) != true) {
		return;
	}
	return;
}

//
// マージン設定ダイアログ
//
void MP_PRINTER::MarginSetup(wxWindow *win)
{
	if (set_margin_page(win) != true) {
		return;
	}
	return;
}

//
// 印刷
//
bool MP_PRINTER::Print(wxWindow *win)
{
	bool rc = true;

	if (text_len == 0) return true;

	/* プリンター選択ダイアログ */
	if (select_printer(win) != true) {
		return true;
	}
#if PRINT_FRAMEWORK == 1
	wxPrinterDC dc(printdata);
	rc = print_data(&dc);
	if (!rc) {
		gErrInfo.SetInfo(__FILE__, __LINE__, myError, myErrPrinting);
		gErrInfo.ShowMsgBox(win);
	}
#endif
	if (draw_mode_bkup) {
		draw_mode = draw_mode_bkup;
		Preview(win);
	}

	return rc;
}

//
// 印刷プレビュー
//
bool MP_PRINTER::PrintPreview(wxWindow *win)
{
	bool rc = true;

	if (text_len == 0) return true;

	wxPrintPreview *preview =
	new wxPrintPreview(new MpPrintout(this, text_file, &pagesetdlg), NULL, &printdlg);
	if (!preview->IsOk()) {
		delete preview;
		gErrInfo.SetInfo(__FILE__, __LINE__, myError, myErrPreview);
		gErrInfo.ShowMsgBox(win);
		return false;
	}

	new MpPreviewFrame(preview, win);

	return rc;
}

//
// プレビュー
//
void MP_PRINTER::Preview(wxWindow *win)
{
	if (text_len == 0 || !draw_mode) return;

//	wxClientDC dc(win);

//	get_device_caps(&dc, &device_info[capsDisplayType]);
	set_paper_size(&device_info[capsDisplayType], rDpPpi.x, rDpPpi.y);

	draw_mode = true;

	// 描画はOnDrawイベントで行う
//	draw_data(&dc, &ddinfo);
}

//
// ズーム
// mag  : %で指定 100=等倍
// 戻り : true
bool MP_PRINTER::Zoom(wxWindow *win, int mag)
{
//	if (draw_mode) {
		rDpPpi.x = (int)((mp_model->GetPPIX()) * mag / 100.0);
		rDpPpi.y = (int)((mp_model->GetPPIY()) * mag / 100.0);
		zoom_mag = mag;

		// サイズを再設定
		set_paper_size(&device_info[capsDisplayType], rDpPpi.x, rDpPpi.y);

		return true;
//	}
//	return false;
}

// プレビュー用画面を描く
void MP_PRINTER::DrawData(wxDC *dc, int left, int top)
{
	if (draw_mode) {
#ifdef USE_CASH_BUFFER
		if (mUseCash) {
			// 描画データをメモリにキャッシュしておく
			if (mDataCashd) {
				draw_from_cashdata(dc, &device_info[capsDisplayType]);
			} else {
				// 高さと幅の計算を行う
				dp_offset.x = 0;
				dp_offset.y = 0;
				draw_data(NULL, &device_info[capsDisplayType]);
				// バッファを作成
				if (create_cashdata(dc)) {
					// バッファに描画
					draw_to_cashdata(dc, &device_info[capsDisplayType]);
				}
			}
		} else
#endif
		{
			dp_offset.x = left;
			dp_offset.y = top;
			draw_data(dc, &device_info[capsDisplayType]);
		}
	}
}
// 印刷用 MpPrintoutクラスより呼ばれる
bool MP_PRINTER::PrintingStart(wxDC *dc, bool preview)
{
	draw_mode_bkup = draw_mode;
	draw_mode = false;

	if (preview) {
		printing = 2;
		dinfo = &device_info[capsPreviewType];
	} else {
		printing = 1;
		dinfo = &device_info[capsPrinterType];
	}
	get_device_caps(dc, dinfo);

	return true;
}
// 印刷用 MpPrintoutクラスより呼ばれる
bool MP_PRINTER::PrintStartDocument(wxDC *dc)
{
	return draw_start_document(dc, dinfo);
}
// 印刷用 MpPrintoutクラスより呼ばれる
bool MP_PRINTER::PrintPage(wxDC *dc, int pageNum)
{
	bool rc = true;
	// 途中のページから印刷する場合でも1ページ目から印刷処理を行う
	for(int page = mPage; page <= pageNum && rc; page++) {
		draw_start_page(dc, dinfo, pageNum);
		draw_page(dc, dinfo, pageNum);
		rc = draw_end_page(dc, dinfo, pageNum);
	}
	return rc;
}
// 印刷用 MpPrintoutクラスより呼ばれる
bool MP_PRINTER::PrintEndDocument(wxDC *dc)
{
	return draw_end_document(dc, dinfo);
}
// 印刷用 MpPrintoutクラスより呼ばれる
bool MP_PRINTER::PrintingEnd(wxDC *dc)
{
	draw_mode = draw_mode_bkup;

	return true;
}

// ファイルを開いてバッファに読み込む
bool MP_PRINTER::ReadFile(wxWindow *win, const wxString &text_path)
{
	wxFileInputStream stream(text_path);
	if (!stream.IsOk()) {
		return false;
	}

	// ファイル読み込み
	memset(text_data, 0, sizeof(text_data));
	stream.Read((void *)text_data,sizeof(text_data));
	// filesize
	text_len = stream.GetSize();

	// ファイル名だけ保存
	wxFileName file_name(text_path);
	text_file = file_name.GetName();

	// 画面に反映
	if (text_len > 0) {
		draw_mode = true;
		mp_model->PushSetting();
	}
	if (win != NULL) {
		Preview(win);
	}

	return true;
}

// ファイルを閉じる
void MP_PRINTER::CloseFile()
{
	ClearData(_T(""));
	release_cashdata();
}

// ドキュメントの幅を返す
int  MP_PRINTER::GetDocWidth()
{
	if (draw_mode) {
		return device_info[capsDisplayType].ph_size.width;
	} else {
		return 0;
	}
}

// ドキュメントの高さを返す
int  MP_PRINTER::GetDocHeight()
{
	if (draw_mode && mp_model != NULL) {
		mp_model->GetDispAspectRatio(&dsr);
		return device_info[capsDisplayType].ph_size.height * dsr.x * mMaxPage / dsr.y;
	} else {
		return 0;
	}
}

// ウィンドウの幅と高さを保存する
void MP_PRINTER::SetWindowSize(int width, int height)
{
	win_size.width = width;
	win_size.height = height;
}

// 印字パラメータを保存
void MP_PRINTER::PushSetting()
{
	mp_model->PushSetting();
}
// 印字パラメータを保存
void MP_PRINTER::PushSetting(int idx)
{
	mp_model_list[idx]->PushSetting();
}
// ページ数を渡す
void MP_PRINTER::GetPageInfo(int *minPage, int *maxPage, int *pageFrom, int *pageTo)
{
	*pageFrom = printdlg.GetFromPage();
	*pageTo = printdlg.GetToPage();
	*minPage = printdlg.GetMinPage();
	*maxPage = printdlg.GetMaxPage();
}

int  MP_PRINTER::GetPPIX(int idx) {
	return mp_model_list[idx]->GetPPIX();
}
int  MP_PRINTER::GetPPIY(int idx) {
	return mp_model_list[idx]->GetPPIY();
}
void MP_PRINTER::SetLineFeedHeight(int idx, double val) {
	mp_model_list[idx]->SetLineFeedHeight(val);
}
double MP_PRINTER::GetLineFeedHeight(int idx) {
	return mp_model_list[idx]->GetLineFeedHeight();
}
void MP_PRINTER::SetLineMode(int idx, int val) {
	mp_model_list[idx]->SetLineMode(val);
}
int  MP_PRINTER::GetLineMode(int idx) {
	return mp_model_list[idx]->GetLineMode();
}
void MP_PRINTER::SetWideMode(int idx, int val) {
	mp_model_list[idx]->SetWideMode(val);
}
int  MP_PRINTER::GetWideMode(int idx) {
	return mp_model_list[idx]->GetWideMode();
}
void MP_PRINTER::SetCharSet(int idx, int val) {
	mp_model_list[idx]->SetCharSet(val);
}
int  MP_PRINTER::GetCharSet(int idx) {
	return mp_model_list[idx]->GetCharSet();
}
void MP_PRINTER::SetCPI(int idx, int val) {
	mp_model_list[idx]->SetCPI(val);
}
int  MP_PRINTER::GetCPI(int idx) {
	return mp_model_list[idx]->GetCPI();
}
wxDC *MP_PRINTER::AttachPrinterDC() {
	if (mPrintDC == NULL) {
		mPrintDC = new wxPrinterDC(pagesetdlg.GetPrintData());
	}
	return mPrintDC;
}
void MP_PRINTER::DetachPrinterDC() {
	if (mPrintDC != NULL) {
		delete mPrintDC;
	}
	mPrintDC = NULL;
}

// 印字用バッファをクリアする
void MP_PRINTER::ClearPrintBuffer()
{
//	memset(pDwBuf,0xff,sizeof(DpType)*MP_PRN_BUF_WIDTH*MP_PRN_BUF_HEIGHT);
	clear_pixeldata(MP_PRN_BUF_WIDTH);
	rDw.width = 0;
	ht_pos = 0;
}

// ドキュメントサイズをセット
void MP_PRINTER::SetDocumentSize(wxPoint *scr_ppi, wxPoint *prn_ppi, wxRect *prect, wxRect *psize)
{
	dinfo->ph_ppi.x = prn_ppi->x;
	dinfo->ph_ppi.y = prn_ppi->y;
	dinfo->ph_size.width = prect->GetWidth();
	dinfo->ph_size.height = prect->GetHeight();
	dinfo->ph_offset.x = - prect->GetLeft();
	dinfo->ph_offset.y = - prect->GetTop();
	dinfo->ph_margin.top = dinfo->margin_mm.top * dinfo->ph_ppi.y * 10 / 254;
	dinfo->ph_margin.bottom = dinfo->margin_mm.bottom * dinfo->ph_ppi.y * 10 / 254;
	dinfo->vm_size.width = dinfo->ph_size.width * dinfo->vm_ppi.x / dinfo->ph_ppi.x;
	dinfo->vm_size.height = dinfo->ph_size.height * dinfo->vm_ppi.y / dinfo->ph_ppi.y;
	dinfo->vm_offset.x = dinfo->ph_offset.x * dinfo->vm_ppi.x / dinfo->ph_ppi.x;
	dinfo->vm_offset.y = dinfo->ph_offset.y * dinfo->vm_ppi.y / dinfo->ph_ppi.y;
	dinfo->vm_margin.top = dinfo->margin_mm.top * dinfo->vm_ppi.y * 10 / 254;
	dinfo->vm_margin.bottom = dinfo->margin_mm.bottom * dinfo->vm_ppi.y * 10 / 254;

#ifdef _DEBUG_LOG
	wxString dbg = _T("SetDocumentSize\n");
	dbg = dbg + _T("  tech: ") + capsTypeStr[dinfo->tech] + _T("\n")
	+ wxString::Format(_T("  margin_mm top: %d mm  bottom: %d mm\n"),dinfo->margin_mm.top, dinfo->margin_mm.bottom)
	+ wxString::Format(_T("  ph_ppi    x: %d  y: %d ppi\n"),dinfo->ph_ppi.x,dinfo->ph_ppi.y)
	+ wxString::Format(_T("  ph_size   w: %d  h: %d px.\n"),dinfo->ph_size.width,dinfo->ph_size.height)
	+ wxString::Format(_T("  ph_offset x: %d  y: %d px.\n"),dinfo->ph_offset.x,dinfo->ph_offset.y)
	+ wxString::Format(_T("  ph_margin top: %d  bottom: %d px.\n"),dinfo->ph_margin.top,dinfo->ph_margin.bottom)
	+ wxString::Format(_T("  vm_ppi    x: %d  y: %d ppi\n"),dinfo->vm_ppi.x,dinfo->vm_ppi.y)
	+ wxString::Format(_T("  vm_size   w: %d  h: %d px.\n"),dinfo->vm_size.width,dinfo->vm_size.height)
	+ wxString::Format(_T("  vm_offset x: %d  y: %d px.\n"),dinfo->vm_offset.x,dinfo->vm_offset.y)
	+ wxString::Format(_T("  vm_margin top: %d  bottom: %d px.\n"),dinfo->vm_margin.top,dinfo->vm_margin.bottom)
	;
	AddDebugLog(dbg);
#endif
}

// 代替印字データをセットする
void MP_PRINTER::SetSubstChar(wxByte ch, size_t len)
{
	for(subst_len=0; subst_len<len && subst_len<1024; subst_len++) {
		subst_data[subst_len]=ch;
	}
	subst_pos=0;
}

//
//
//
//
//
// 描画したデータを保持しておく
bool MP_PRINTER::create_cashdata(wxDC *dc)
{
#ifdef USE_CASH_BUFFER
	release_cashdata();

	bmpCashd = new wxBitmap(GetDocWidth(), GetDocHeight(), dc->GetDepth());
	if (!bmpCashd->IsOk()) {
		return false;
	}
	dcCashd = new wxMemoryDC(*bmpCashd);
	if (!dcCashd->IsOk()) {
		delete bmpCashd;
		return false;
	}

	mDataCashd = true;
#endif
	return true;
}
// 保持しておくバッファを破棄
void MP_PRINTER::release_cashdata()
{
#ifdef USE_CASH_BUFFER
	if (mDataCashd) {
		delete dcCashd;
		delete bmpCashd;
		dcCashd = NULL;
		bmpCashd = NULL;
	}
	mDataCashd = false;
#endif
}

// プリンターなどのデバイス情報を得る
void MP_PRINTER::get_device_caps(wxDC *dc, CapsInfoType *inf)
{
	// not used
	// ミリメートル（mm）単位の画面の物理的な幅、高さ。
	wxSize msize = dc->GetSizeMM();
	inf->size_mm.width = msize.GetWidth();
	inf->size_mm.height = msize.GetHeight();

	// 論理インチ当たりのピクセル数。
	wxSize ppi = dc->GetPPI();
	inf->ph_ppi.x = ppi.GetWidth();
	inf->ph_ppi.y = ppi.GetHeight();

	if (inf->tech == capsPrinterType) {
		// 印刷デバイス用
		wxPrinterDC *pdc = (wxPrinterDC *)dc;
		wxRect prect = pdc->GetPaperRect();
		// ページの物理的な幅をデバイス単位で表します。たとえば、8.5"×11" の用紙に 600dpi の解像度で印刷するよう設定されているプリンタでは、物理的な幅は 5,100 デバイス単位になります（8.5×600=5,100）。ほとんどの場合、用紙の印字可能領域は用紙の物理的な幅よりも多少小さいこと、そして印字可能領域は物理的な幅を決して上回らないことに注意してください。
		inf->ph_size.width = prect.GetWidth();
		// ページの物理的な高さをデバイス単位で表します。たとえば、8.5"×11" の用紙に 600dpi の解像度で印刷するよう設定されているプリンタでは、物理的な高さは 6,600 デバイス単位になります（11×600=6,600）。ほとんどの場合、用紙の印字可能領域は用紙の物理的な高さよりも多少小さいこと、そして印字可能領域は物理的な高さを決して上回らないことに注意してください。
		inf->ph_size.height = prect.GetHeight();
		// 物理的なページの左端から印刷可能領域の左端までの距離をデバイス単位で表します。
		inf->ph_offset.x = - prect.GetLeft();
		// 物理的なページの上端から印刷可能領域の上端までの距離をデバイス単位で表します。
		inf->ph_offset.y = - prect.GetTop();

		inf->vm_size.width  = inf->ph_size.width * inf->vm_ppi.x / inf->ph_ppi.x;
		inf->vm_size.height = inf->ph_size.height * inf->vm_ppi.y / inf->ph_ppi.y;
		inf->vm_offset.x = inf->ph_offset.x * inf->vm_ppi.x / inf->ph_ppi.x;
		inf->vm_offset.y = inf->ph_offset.y * inf->vm_ppi.y / inf->ph_ppi.y;
	}

	// 余白(pixcel単位)
	inf->ph_margin.top = (inf->margin_mm.top * inf->ph_ppi.y * 10 / 254);
	inf->ph_margin.bottom = (inf->margin_mm.bottom * inf->ph_ppi.y * 10 / 254);
	inf->vm_margin.top    = inf->margin_mm.top * inf->vm_ppi.y * 10 / 254;
	inf->vm_margin.bottom = inf->margin_mm.bottom * inf->vm_ppi.y * 10 / 254;

#ifdef _DEBUG_LOG
	wxString dbg = _T("get_device_caps\n");
	dbg = dbg + _T("  tech: ") + capsTypeStr[inf->tech] + _T("\n")
	+ wxString::Format(_T("  size_mm width: %d mm  height: %d mm\n"),inf->size_mm.width, inf->size_mm.height)
	+ wxString::Format(_T("  margin_mm top: %d mm  bottom: %d mm\n"),inf->margin_mm.top, inf->margin_mm.bottom)
	+ wxString::Format(_T("  ph_ppi    x: %d  y: %d ppi\n"),inf->ph_ppi.x,inf->ph_ppi.y)
	+ wxString::Format(_T("  ph_size   w: %d  h: %d px.\n"),inf->ph_size.width,inf->ph_size.height)
	+ wxString::Format(_T("  ph_offset x: %d  y: %d px.\n"),inf->ph_offset.x,inf->ph_offset.y)
	+ wxString::Format(_T("  ph_margin top: %d  bottom: %d px.\n"),inf->ph_margin.top,inf->ph_margin.bottom)
	+ wxString::Format(_T("  vm_ppi    x: %d  y: %d ppi\n"),inf->vm_ppi.x,inf->vm_ppi.y)
	+ wxString::Format(_T("  vm_size   w: %d  h: %d px.\n"),inf->vm_size.width,inf->vm_size.height)
	+ wxString::Format(_T("  vm_offset x: %d  y: %d px.\n"),inf->vm_offset.x,inf->vm_offset.y)
	+ wxString::Format(_T("  vm_margin top: %d  bottom: %d px.\n"),inf->vm_margin.top,inf->vm_margin.bottom)
	+ wxString::Format(_T("  win_size w: %d  h: %d px.\n"),win_size.width,win_size.height)
	;
	AddDebugLog(dbg);
#endif
}

// 画面用に仮想で紙サイズを設定
void MP_PRINTER::set_paper_size(CapsInfoType *inf, int ppix, int ppiy)
{
#ifdef _DEBUG_LOG
	wxString dbg = _T("set_paper_size");
	dbg += _T("  tech: ") + capsTypeStr[inf->tech];
	AddDebugLog(dbg);
#endif
	// 解像度を指定（拡大・縮小率を計算した値をセット）
	inf->ph_ppi.x = ppix;
	inf->ph_ppi.y = ppiy;
	// 画面の場合はDCの左上が原点となるのでオフセットは常に0にする
	inf->ph_offset.x = 0;
	inf->ph_offset.y = 0;
	inf->vm_offset.x = 0;
	inf->vm_offset.y = 0;

	wxPrinterDC pdc(printdata);
#ifdef __WXGTK__
	// TODO: GTKではGetPaperRectに失敗する。
	if (0) {
#else
	if (pdc.IsOk()) {
#endif
		// プリンター情報から計算
		wxSize msize = pdc.GetSizeMM();
		wxRect pdc_rect = pdc.GetPaperRect();
		wxSize pdc_size = pdc.GetSize();
		wxSize pdc_ppi = pdc.GetPPI();
#ifdef _DEBUG_LOG
		dbg = _T("set_paper_size --- get from printerDC\n");
		dbg += wxString::Format(_T("  pdc_rect left: %d  top: %d px.\n"),pdc_rect.GetLeft(),pdc_rect.GetTop())
		+ wxString::Format(_T("  pdc_rect w: %d  h: %d px.\n"),pdc_rect.GetWidth(),pdc_rect.GetHeight())
		+ wxString::Format(_T("  pdc_size w: %d  h: %d px.\n"),pdc_size.GetWidth(),pdc_size.GetHeight())
		+ wxString::Format(_T("  pdc_ppi  x: %d  y: %d ppi"),pdc_ppi.GetX(),pdc_ppi.GetY());
		AddDebugLog(dbg);
#endif
		// 設定がなければA4縦
		if (pdc_rect.GetWidth() <= 0) pdc_rect.SetWidth(2100 * pdc_ppi.GetX() / 254);
		if (pdc_rect.GetHeight() <= 0) pdc_rect.SetHeight(2970 * pdc_ppi.GetY() / 254);

		inf->size_mm.width  = msize.GetWidth();
		inf->size_mm.height = msize.GetHeight();

		inf->ph_size.width  = pdc_rect.GetWidth() * inf->ph_ppi.x / pdc_ppi.GetX();
		inf->ph_size.height = pdc_rect.GetHeight() * inf->ph_ppi.y / pdc_ppi.GetY();

		inf->vm_size.width  = pdc_rect.GetWidth() * inf->vm_ppi.x / pdc_ppi.GetX();
		inf->vm_size.height = pdc_rect.GetHeight() * inf->vm_ppi.y / pdc_ppi.GetY();

#if 0
		// 余白はプリンターのサイズから計算する
		inf->ph_margin.top    = - pdc_rect.GetTop();
		inf->ph_margin.bottom = pdc_rect.GetHeight() - pdc_size.GetHeight() + pdc_rect.GetTop();
		inf->margin_mm.top    = (inf->ph_margin.top * 254 / pdc_ppi.GetY() / 10);
		inf->margin_mm.bottom = (inf->ph_margin.bottom * 254 / pdc_ppi.GetY() / 10);
		vm_margin.top    = inf->ph_margin.top * inf->vm_ppi.x / pdc_ppi.GetX();
		vm_margin.bottom = inf->ph_margin.bottom * inf->vm_ppi.y / pdc_ppi.GetY();
		inf->ph_margin.top    = inf->ph_margin.top * inf->ph_ppi.y / pdc_ppi.GetY();
		inf->ph_margin.bottom = inf->ph_margin.bottom * inf->ph_ppi.y / pdc_ppi.GetY();
#else
		inf->vm_margin.top    = inf->margin_mm.top * inf->vm_ppi.y * 10 / 254;
		inf->vm_margin.bottom = inf->margin_mm.bottom * inf->vm_ppi.y * 10 / 254;
		inf->ph_margin.top    = inf->margin_mm.top * inf->ph_ppi.y * 10 / 254;
		inf->ph_margin.bottom = inf->margin_mm.bottom * inf->ph_ppi.y * 10 / 254;
#endif
	} else {
		// ページ設定の紙サイズに合わせる
		wxSize psize = pagesetdlg.GetPaperSize();
#ifdef _DEBUG_LOG
		dbg = _T("set_paper_size -- get from pagesetdlg\n");
		dbg += wxString::Format(_T("  paper_size w: %d  h: %d"),psize.GetWidth(),psize.GetHeight());
		AddDebugLog(dbg);
#endif
		// 設定がなければA4縦
		if (psize.GetWidth() <= 0) psize.SetWidth(210);
		if (psize.GetHeight() <= 0) psize.SetHeight(297);

		if (pagesetdlg.GetPrintData().GetOrientation() == wxLANDSCAPE) {
			// 横
			int w = psize.GetWidth();
			psize.SetWidth(psize.GetHeight());
			psize.SetHeight(w);
		}
		inf->size_mm.width  = psize.GetWidth();
		inf->size_mm.height = psize.GetHeight();

		inf->ph_size.width  = psize.GetWidth() * inf->ph_ppi.x * 10 / 254;
		inf->ph_size.height = psize.GetHeight() * inf->ph_ppi.y * 10 / 254;
		inf->vm_size.width  = psize.GetWidth() * inf->vm_ppi.x * 10 / 254;
		inf->vm_size.height = psize.GetHeight() * inf->vm_ppi.y * 10 / 254;

		inf->ph_margin.top    = inf->margin_mm.top * inf->ph_ppi.y * 10 / 254;
		inf->ph_margin.bottom = inf->margin_mm.bottom * inf->ph_ppi.y * 10 / 254;
		inf->vm_margin.top    = inf->margin_mm.top * inf->vm_ppi.y * 10 / 254;
		inf->vm_margin.bottom = inf->margin_mm.bottom * inf->vm_ppi.y * 10 / 254;
	}

	// バッファを再作成
#ifdef USE_PAGE_BUFFER
	delete dcPage;
	wxBitmap bmpPage(inf->ph_size.width, inf->ph_size.height, 24);
	if (!bmpPage.IsOk()) {
		gErrInfo.SetInfo(__FILE__, __LINE__, myError, myErrSystem, _T("Cannot create bmpPage."));
		gErrInfo.ShowMsgBox();
		return;
	}
	dcPage = new wxMemoryDC(bmpPage);
	dcPage->Clear();
#endif

#ifdef _DEBUG_LOG
	dbg = _T("")
	+ wxString::Format(_T("  size_mm width: %d mm  height: %d mm\n"),inf->size_mm.width, inf->size_mm.height)
	+ wxString::Format(_T("  margin_mm top: %d  bottom: %d mm\n"),inf->margin_mm.top,inf->margin_mm.bottom)
	+ wxString::Format(_T("  ph_ppi    x: %d  y: %d ppi\n"),inf->ph_ppi.x,inf->ph_ppi.y)
	+ wxString::Format(_T("  ph_size   w: %d  h: %d px.\n"),inf->ph_size.width,inf->ph_size.height)
	+ wxString::Format(_T("  ph_offset x: %d  y: %d px.\n"),inf->ph_offset.x,inf->ph_offset.y)
	+ wxString::Format(_T("  ph_margin top: %d  bottom: %d px.\n"),inf->ph_margin.top,inf->ph_margin.bottom)
	+ wxString::Format(_T("  vm_ppi    x: %d  y: %d ppi\n"),inf->vm_ppi.x,inf->vm_ppi.y)
	+ wxString::Format(_T("  vm_size   w: %d  h: %d px.\n"),inf->vm_size.width,inf->vm_size.height)
	+ wxString::Format(_T("  vm_offset x: %d  y: %d px.\n"),inf->vm_offset.x,inf->vm_offset.y)
	+ wxString::Format(_T("  vm_margin top: %d  bottom: %d px.\n"),inf->vm_margin.top,inf->vm_margin.bottom)
	;
	AddDebugLog(dbg);
#endif
}

//
// プリンター選択ダイアログ を表示
//
//
bool MP_PRINTER::select_printer(wxWindow *win)
{
	/* プリンター選択ダイアログ */
#ifdef __WXGTK__
	printdlg.SetSelection(false);
#else
	printdlg.SetSelection(true);
#endif
	printdlg.SetNoCopies(1);
	printdlg.SetFromPage(1);
	printdlg.SetToPage(mMaxPage);
	printdlg.SetMinPage(1);
	printdlg.SetMaxPage(mMaxPage);

#if PRINT_FRAMEWORK == 1
	// TODO: MacでMyPrinterを使用するとStretchBlitに失敗して印刷できないので
	// こちらを使用する
#if 1
	wxPrintDialog dlg(win, &printdlg);
	if (dlg.ShowModal() != wxID_OK) {
		return false;
	}
	printdlg = dlg.GetPrintDialogData();
	printdata = printdlg.GetPrintData();
	pagesetdlg.SetPrintData(printdata);
#else
	wxPrinter printer(&printdlg);
	wxDC *pdc = printer.PrintDialog(win);
	if (pdc == NULL) {
        if (wxPrinter::GetLastError() == wxPRINTER_ERROR) {
			gErrInfo.SetInfo(__FILE__, __LINE__, myError, myErrPrinting);
			gErrInfo.ShowMsgBox(win);
        }
		return false;
	}
	printdlg = printer.GetPrintDialogData();
	printdata = printdlg.GetPrintData();
	pagesetdlg.SetPrintData(printdata);
#endif
#else
	wxPrinter printer(&printdlg);
	MpPrintout printout(this, text_file, &pagesetdlg);
    if (!printer.Print(win, &printout, true /*prompt*/)) {
        if (wxPrinter::GetLastError() == wxPRINTER_ERROR) {
			gErrInfo.SetInfo(__FILE__, __LINE__, myError, myErrPrinting);
			gErrInfo.ShowMsgBox(win);
        }
		// TODO: Winでは常にここに来るのだが。。。
//		return false;
	}
	printdlg = printer.GetPrintDialogData();
	printdata = printdlg.GetPrintData();
	pagesetdlg.SetPrintData(printdata);
#endif

	return true;
}

//
// プリンターページ設定ダイアログ を表示
//
//
bool MP_PRINTER::set_printer_page(wxWindow *win)
{
	/* プリンターページ設定ダイアログ */
	pagesetdlg.SetDefaultMinMargins(true);

	wxPageSetupDialog dlg(win, &pagesetdlg);
	if (dlg.ShowModal() != wxID_OK) {
		return false;
	}
	pagesetdlg = dlg.GetPageSetupDialogData();

	printdata = pagesetdlg.GetPrintData();
	printdlg.SetPrintData(printdata);

	wxPoint pMarginTopLeft = pagesetdlg.GetMarginTopLeft();
	wxPoint pMarginBottomRight = pagesetdlg.GetMarginBottomRight();
	for(int i=0; i<capsUnknownType; i++) {
		device_info[i].margin_mm.top = pMarginTopLeft.y;
		device_info[i].margin_mm.bottom = pMarginBottomRight.y;
	}

#ifdef _DEBUG_LOG
	wxSize sz = pagesetdlg.GetPaperSize();
	wxString dbg(_T("set_printer_page\n"));
	dbg += wxString::Format(_T("  PaperId: %d\n"),pagesetdlg.GetPaperId());
	dbg += wxString::Format(_T("  Size : w:%d h:%d mm\n")
		,sz.GetWidth(),sz.GetHeight());
	dbg += wxString::Format(_T("  Mergin: t:%d b:%d l:%d r:%d mm\n")
		,pMarginTopLeft.y,pMarginBottomRight.y,pMarginTopLeft.x,pMarginBottomRight.x);
	dbg += _T("  Printer: ") + printdata.GetPrinterName() + _T("\n");
	AddDebugLog(dbg);
#endif

	// 設定した紙サイズに合わせて画面も再設定
	set_paper_size(&device_info[capsDisplayType]
		, device_info[capsDisplayType].ph_ppi.x, device_info[capsDisplayType].ph_ppi.y);

	release_cashdata();
	return true;
}

//
// マージン設定ダイアログ を表示
//
//
bool MP_PRINTER::set_margin_page(wxWindow *win)
{
	pagesetdlg.SetDefaultMinMargins(true);

	MarginBox dlg(win, wxID_ANY, &pagesetdlg);
	if (dlg.ShowModal() != wxID_OK) {
		return false;
	}
	printdata = pagesetdlg.GetPrintData();
	printdlg.SetPrintData(printdata);

	wxPoint pMarginTopLeft = pagesetdlg.GetMarginTopLeft();
	wxPoint pMarginBottomRight = pagesetdlg.GetMarginBottomRight();
	for(int i=0; i<capsUnknownType; i++) {
		device_info[i].margin_mm.top = pMarginTopLeft.y;
		device_info[i].margin_mm.bottom = pMarginBottomRight.y;
	}

	// 設定した紙サイズに合わせて画面も再設定
	set_paper_size(&device_info[capsDisplayType]
		, device_info[capsDisplayType].ph_ppi.x, device_info[capsDisplayType].ph_ppi.y);

	release_cashdata();
	return true;
}

//
// 制御コードの解析
//
//
int MP_PRINTER::parse_ctrl_code(wxByte *data)
{
	int bytes = 0;

	// 連続改行中や代替データ出力中はコードの解析をしない
	if (line_feed_number > 0 || subst_len > 0) {
		return bytes;
	}
	// 制御コード（機種依存）
	bytes = mp_model->ParseCtrlCode(data, &rDw.width, &rDw.height, &feed);
	if (bytes != 0) {
		rDwSz.width = (mp_model->GetLineWidth());	// 1行の幅
		line_feed_height = mp_model->GetLineFeedHeight();	// 改行幅
		paper_feed_height = mp_model->GetPaperFeedHeight();	// 紙送り幅
	}
	return bytes;
}

// データを描画する
//
//
bool MP_PRINTER::draw_data(wxDC *dc, CapsInfoType *inf)
{
#ifdef USE_PAGE_BUFFER
	draw_start_document(dcPage, inf);
	for(int page = 1; page < 100 ; page++) {
		draw_start_page(dcPage, inf, page);
		draw_page(dcPage, inf, page);
		if (dc != NULL) {
			dc->Blit(win_offset.x, inf->ph_size.height * (page - 1), inf->ph_size.width, inf->ph_size.height, dcPage, 0, 0);
		}
		if (!draw_end_page(dc, inf, page)) {
			break;
		}
	}
#else
	draw_start_document(dc, inf);
	for(int page = 1; page < 100 ; page++) {
		draw_start_page(dc, inf, page);
		draw_page(dc, inf, page);
		if (!draw_end_page(dc, inf, page)) {
			break;
		}
	}
#endif
	draw_end_document(dc, inf);
	return true;
}

// 印刷する
//
//
bool MP_PRINTER::print_data(wxDC *dc)
{
	PrintingStart(dc, false);
	PrintStartDocument(dc);
	for(int page = printdlg.GetFromPage(); page <= printdlg.GetToPage(); page++) {
		PrintPage(dc, page);
	}
	PrintEndDocument(dc);
	PrintingEnd(dc);
	return true;
}

// ドキュメント描画開始時処理
bool MP_PRINTER::draw_start_document(wxDC *dc, CapsInfoType *inf)
{
	// 画面上の縦横比
	mp_model->GetDispAspectRatio(&dsr);
	// 開始時のパラメータをセット
	mp_model->PopSetting();

	rDw.left = 0;					// 描画開始位置・左端
	rDw.top = 0;
	rDw.width = 0;					// 描画用バッファに書かれたデータのwidth
	rDw.height = MP_PRN_BUF_HEIGHT;	// 描画用バッファに書かれたデータのheight

	rDwSz.width = (mp_model->GetLineWidth());	// 1行の幅
	rDwSz.height = MP_PRN_BUF_HEIGHT;			// 1行の高さ


	// 印刷範囲は拡大縮小を考慮した値となる->Dp
	// 印刷範囲は拡大縮小を考慮した値とならない->Vm
	// 印刷範囲の最大幅
//	rDpMax.width = (MP_PRN_BUF_WIDTH * inf->ph_ppi.x / inf->vm_ppi.x);
//	rDpMax.height = 0;
	rVmMax.width = MP_PRN_BUF_WIDTH;
	rVmMax.height = 0;

	// 印刷範囲左端
//	rDp.left = (inf->ph_size.width - rDpMax.width) / 2 - inf->ph_offset.x;
	rVm.left = (inf->vm_size.width - rVmMax.width) / 2 - inf->vm_offset.x;
	// 印刷範囲右端
//	rDp.right = rDp.left + rDpMax.width;
	rVm.right = rVm.left + rVmMax.width;
	// 印刷範囲上端
//	rDp.top = inf->ph_margin.top - inf->ph_offset.y;
	rVm.top = inf->vm_margin.top - inf->vm_offset.y;
//	if (rDp.top < 0) rDp.top = 0;
	if (rVm.top < 0) rVm.top = 0;
	// 印刷範囲下端
//	rDp.bottom = (calc_dp_bottom(inf, 0));
	rVm.bottom = (calc_vm_bottom(inf, 0));

//	dp_y = rDp.top;		// 印字行(ページ内)
//	dp_ry = rDp.top;	// 印字行(画面へ出力する場合、上端からの累積座標となる)
	vm_y = rVm.top;		// 印字行(ページ内)
	vm_ry = rVm.top;	// 印字行(画面へ出力する場合、上端からの累積座標となる)

	rMx.left = 0;	// 合成用バッファに書かれたデータの左端
	rMx.top = 0;	// 合成用バッファに書かれたデータの上端 常に0
	rMx.right = 0;	// 合成用バッファに書かれたデータの右端(最大値)
	rMx.bottom = 0;	// 合成用バッファに書かれたデータの下端

	// 画面表示で描画処理を行う範囲
//	rDwIn.left = 0;
//	rDwIn.top = (-48 * inf->ph_ppi.y / inf->vm_ppi.y) - inf->ph_margin.top - inf->ph_margin.bottom + dp_offset.y;
//	rDwIn.right = win_size.width;
//	rDwIn.bottom = win_size.height + dp_offset.y;
	rVmIn.left = 0;
	rVmIn.top = -48 - inf->vm_margin.top - inf->vm_margin.bottom + (dp_offset.y * inf->vm_ppi.y / inf->ph_ppi.y);
	rVmIn.right = win_size.width * inf->vm_ppi.y / inf->ph_ppi.y;
	rVmIn.bottom = (win_size.height + dp_offset.y) * inf->vm_ppi.y / inf->ph_ppi.y;
	rVmIn.top = rVmIn.top * dsr.y / dsr.x;
	rVmIn.bottom = rVmIn.bottom * dsr.y / dsr.x;

	// 画面内か
	draw_win = false;
	// 印刷対象ページか
	print_page = false;
	// 紙送りフラグ
	feed = 0;
	// 改行数
	line_feed_number = 0;
	// 垂直、水平タブ位置初期化
	vt_line_number = 0;
	ht_pos = 0;
	// 改ページ指示
	change_page = false;
	// ページ数
	mPage = 1;
	// ページ終りか
	endPage = false;
	// 代替テキスト
	subst_len = 0;
	memset(subst_data, 0, sizeof(subst_data));

	if (inf->tech == capsDisplayType) {
		// ラスタディスプレイの場合
		if (dc != NULL) {
			// 画面に描画する場合の描画位置
#ifdef USE_CASH_BUFFER
			if (mUseCash) {
				win_offset.x = 0;
			} else
#endif
			{
				win_offset.x = (win_size.width - inf->ph_size.width) / 2;
				if (win_offset.x < 0) win_offset.x = 0;
#ifndef USE_PAGE_BUFFER
//				rDp.left += win_offset.x;
//				rDp.right += win_offset.x;
				rVm.left += (win_offset.x * inf->vm_ppi.x / inf->ph_ppi.x);
				rVm.right += (win_offset.x * inf->vm_ppi.x / inf->ph_ppi.x);
#endif
			}

			// 描画用ペンの設定
			dc->SetPen(wxPen(wxColour(0, 0, 0),1));
			// 描画用フォントの設定
			set_sys_font(dc, inf->ph_ppi.x / 5, inf->ph_ppi.x / 3);
		}

		// 画面内か
//		draw_win = is_inwindow(dc, inf, dp_ry);
		draw_win = is_invmwindow(dc, inf, vm_ry);

		print_page = true;

	} else if (inf->tech == capsPrinterType) {
		// ラスタプリンタの場合
		draw_win = true;

		// 印刷開始
		dc->StartDoc(text_file);

	} else if (inf->tech == capsPreviewType) {
		// 印刷プレビューの場合
		draw_win = true;

	}

	if (draw_win) {
		clear_mixbuffer(MP_PRN_BUF_WIDTH);
		clear_pixeldata(MP_PRN_BUF_WIDTH);
	}
	line_feed_height = mp_model->GetLineFeedHeight();
	paper_feed_height = mp_model->GetPaperFeedHeight();

	text_pos = 0;

	return true;
}

// ページ開始時処理
bool MP_PRINTER::draw_start_page(wxDC *dc, CapsInfoType *inf, int target_page)
{
	if (inf->tech == capsDisplayType) {
		// ラスタディスプレイの場合
//		dp_ry = inf->ph_size.height * (mPage - 1) + rDp.top;	// 先頭
		vm_ry = inf->vm_size.height * (mPage - 1) + rVm.top;	// 先頭
		// 画面内か
//		draw_win = is_inwindow(dc, inf, dp_ry);
		draw_win = is_invmwindow(dc, inf, vm_ry);

		// ページ番号＆枠の表示
		if (dc != NULL) {
			print_page_no(dc, inf, mPage);
		}
	} else if (inf->tech == capsPrinterType) {
		// ラスタプリンタの場合
//		dp_ry = rDp.top;	// 先頭
		vm_ry = rVm.top;	// 先頭

		if (target_page == mPage) {
			print_page = true;
			dc->StartPage();
		}
	} else if (inf->tech == capsPreviewType) {
		// 印刷プレビューの場合
//		dp_ry = rDp.top;	// 先頭
		vm_ry = rVm.top;	// 先頭

		if (target_page == mPage) {
			print_page = true;
		}
	}
//	dp_y = rDp.top;	// 先頭
	vm_y = rVm.top;	// 先頭
	change_page = false;

	return true;
}

// ページ描画処理
bool MP_PRINTER::draw_page(wxDC *dc, CapsInfoType *inf, int target_page)
{
//	AddDebugLog(wxString::Format(_T("draw_page: %d"), mPage));

	while(text_pos < text_len) {
		// 制御コードか？
		int step = parse_ctrl_code(&text_data[text_pos]);
		if (step > 0) {
			// 制御コードの場合は印字スキップ
			text_pos += step - 1;
			if (feed == 0) {
				text_pos++;
				continue;
			}
		}
		// バッファがいっぱいになった
		if (rDw.width >= rDwSz.width && (feed & MP_BIT_CR) == 0) {
			feed |= MP_BIT_LF;	// 改行を指示
			if (line_feed_number == 0) line_feed_number = 1;
		}
		// バッファがいっぱいになった or 紙送り/改行
		if (feed != 0) {
			// 画面内であれば描画処理
			put_buffer(dc, inf, draw_win && print_page, false);

			// 復帰
			if ((feed & ~MP_BIT_PF) != 0) {
				rDw.width = (rDw.width - rDwSz.width);
				if (rDw.width < 0) {
					rDw.width = 0;
				}
				ht_pos = 0;

				// 画面内であれば描画処理
				if (draw_win) {
					// 描画用バッファも復帰にあわせてクリア
					shift_pixeldata(rDwSz.width);
				}
			} else {
				// 画面内であれば描画処理
				if (draw_win) {
					// 描画用バッファ印字した分はクリア
					clear_pixeldata(rDwSz.width);
				}
			}

			// 改行
			if ((feed & ~MP_BIT_CR) != 0) {
				// 改行幅
//				double next_line;
				double vm_next_line;
				if ((feed & MP_BIT_PF) == 0) {
					// 改行
//					next_line = (double)(line_feed_height * inf->ph_ppi.y / inf->vm_ppi.y);
					vm_next_line = line_feed_height;
					vt_line_number++;
				} else {
					// 紙送り
//					next_line = (double)(paper_feed_height * inf->ph_ppi.y / inf->vm_ppi.y);
					vm_next_line = paper_feed_height;
					paper_feed_height = 0.0;
					mp_model->SetPaperFeedHeight(0.0);
				}
//				dp_y += next_line;
//				dp_ry += next_line;
				vm_y_prev = vm_y;
				vm_y += vm_next_line;
				vm_ry_prev = vm_ry;
				vm_ry += vm_next_line;

				// 印刷範囲下端を再計算
//				rDp.bottom = calc_dp_bottom(inf, rMx.bottom);
				rVm.bottom = calc_vm_bottom(inf, rMx.bottom);

				// 合成バッファもクリア
//				if (rMx.bottom >= rMxMaxV.bottom) {
//					if (draw_win) {
//						clear_mixbuffer(MP_PRN_BUF_WIDTH);
//					}
//				} else {
					// MP_PRN_BUF_HEIGHT pixcel未満では印字が重なるように現在行の一部を残す
					if (draw_win) {
						shift_mixbuffer(MP_PRN_BUF_WIDTH, (int)(vm_ry - vm_ry_prev));
					}
//				}

				// 画面内か
//				draw_win = is_inwindow(dc, inf, dp_ry);
				draw_win = is_invmwindow(dc, inf, vm_ry);

			}

			// 改行or復帰時のイベント
			mp_model->OnLineFeed();

			// 印刷範囲の下端に達した場合、改ページ指示
//			if ((feed & MP_BIT_FF) || (dp_y > rDp.bottom && rDp.bottom > 0)) {
			if ((feed & MP_BIT_FF) || (vm_y > rVm.bottom && rVm.bottom > 0)) {
				change_page = true;
				// 次ページの先頭位置を計算
				vm_ry = inf->vm_size.height * mPage + rVm.top;
				// 画面内か
				draw_win = is_invmwindow(dc, inf, vm_ry);
			}
		} // end of 改行処理

		//
		// 印字データを描画エリアに出力
		//
		if (subst_len > 0) {
			// 代替データを出力
			mp_model->PutData(&subst_data[subst_pos++], bufDw, &rDw.width, &rDw.height, draw_win);
			ht_pos++;
			if (subst_pos >= subst_len) {
				subst_len = 0;
			}
		} else if (line_feed_number > 0) {
			// 連続改行中は印字データの出力はしない
			line_feed_number--;
			if (line_feed_number == 0) text_pos++;
		} else {
			mp_model->PutData(&text_data[text_pos++], bufDw, &rDw.width, &rDw.height, draw_win);
			ht_pos++;
		}

		// 改ページ
		if (change_page == true && (rDw.width > 0 || (feed & MP_BIT_LF))) {
			break;
		}
	}
	// 最終ページの判断：
	// 改ページ指示があったがその後ろに印字するデータがない場合
	// 印字するデータがなくなった場合
	if ((change_page == true && rDw.width == 0 && (feed & MP_BIT_LF) == 0) || (change_page == false && text_pos >= text_len)) {
		endPage = true;

		// バッファの残りを出力
		if (rDw.width > 0 || rMx.right > rMx.left) {
			put_buffer(dc, inf, draw_win && print_page, true);
		}
		// 改行or復帰時のイベント
		mp_model->OnLineFeed();

		return false;
	}
	return true;
}

// ページ終了時処理
bool MP_PRINTER::draw_end_page(wxDC *dc, CapsInfoType *inf, int target_page)
{
	if (inf->tech == capsDisplayType) {
		// ラスタディスプレイの場合

	} else if (inf->tech == capsPrinterType) {
		// ラスタプリンタの場合
		// 改ページ
		if (target_page == mPage) {
			dc->EndPage();
		}
		print_page = false;

	} else if (inf->tech == capsPreviewType) {
		// 印刷プレビューの場合
		print_page = false;
	}
	// 最終ページの場合
	if (endPage) {
		// 印字終わり
		return false;
	}

	mPage++;

	return true;
}

// ドキュメント描画終了時処理
bool MP_PRINTER::draw_end_document(wxDC *dc, CapsInfoType *inf)
{
	if (inf->tech == capsDisplayType) {
		// ラスタディスプレイの場合
		// ページ数を設定
		printdlg.SetNoCopies(1);
		printdlg.SetFromPage(1);
		printdlg.SetToPage(mPage);
		printdlg.SetMinPage(1);
		printdlg.SetMaxPage(mPage);
		mMaxPage = mPage;

	} else if (inf->tech == capsPrinterType) {
		// ラスタプリンタの場合
		// 印刷終了
		dc->EndDoc();
	} else if (inf->tech == capsPreviewType) {
		// 印刷プレビューの場合

	}

	return true;
}

// 画面内か
bool MP_PRINTER::is_inwindow(wxDC *dc, CapsInfoType *inf, double y)
{
	bool rc = true;

	if (dc == NULL) {
		return false;
	}
#ifdef USE_CASH_BUFFER
	if (mUseCash) {
	} else
#endif
	{
		if (inf->tech == capsDisplayType) {
			// ラスタディスプレイの場合
			// 画面内か
//			int draw_winy = (int)(y * dsr.x / dsr.y);
//			rc = (draw_winy >= rDwIn.top && draw_winy <= rDwIn.bottom);
		}
	}
	return rc;
}
bool MP_PRINTER::is_invmwindow(wxDC *dc, CapsInfoType *inf, double y)
{
	bool rc = true;

	if (dc == NULL) {
		return false;
	}
#ifdef USE_CASH_BUFFER
	if (mUseCash) {
	} else
#endif
	{
		if (inf->tech == capsDisplayType) {
			// ラスタディスプレイの場合
			// 画面内か
			int draw_winy = (int)y;
			rc = (draw_winy >= rVmIn.top && draw_winy <= rVmIn.bottom);
		}
	}
	return rc;
}

// バッファの内容をデバイスに出力
void MP_PRINTER::put_buffer(wxDC *dc, CapsInfoType *inf, bool draw, bool last)
{
	MyRect dp = {0,0,0,0};
	int dw_width;

	// 合成する範囲の決定
	rDw.left = (MP_PRN_BUF_WIDTH - rDwSz.width) / 2;
	dw_width = wxMin(rDw.width, rDwSz.width);
	// 合成
	if (draw && rDw.width > 0 && rDw.height > 0) {
		copy_pixeldata(rDw.left, rDw.top, dw_width, rDw.height, 0, 0);
	}

	// 出力範囲の決定
	// 一番下端を決定
//	if (paper_feed_height <= 0) {
//		rMx.bottom = (int)(dp_ry + line_feed_height) - (int)dp_ry;
//		rMx.bottom = (int)(vm_ry + line_feed_height) - (int)vm_ry;
//	} else {
//		rMx.bottom = (int)(dp_ry + paper_feed_height) - (int)dp_ry;
//		rMx.bottom = (int)(vm_ry + paper_feed_height) - (int)vm_ry;
//	}
	rMx.bottom = rDw.height;
	if (last || rMx.bottom > MP_PRN_BUF_HEIGHT) {
		rMx.bottom = MP_PRN_BUF_HEIGHT;
	}
	rMxMaxV.bottom = rMx.bottom;

	// 一番左端と右端を決定
	rMx.left = MP_PRN_BUF_WIDTH;
	rMx.right = 0;
	for(int i=0; i<rMx.bottom; i++) {
		if (rDw.left < rMxMax[i].left) {
			rMxMax[i].left = rDw.left;
		}
		if (rMxMax[i].left < rMx.left) {
			rMx.left = rMxMax[i].left;
		}
		if (rMxMax[i].right < (rDw.left + rDw.width)) {
			rMxMax[i].right =  rDw.left + rDw.width;
		}
		if (rMx.right < rMxMax[i].right) {
			rMx.right = rMxMax[i].right;
		}
	}

	// 出力
	if (draw) {
		if (inf->tech == capsDisplayType) {
			// 画面描画
//			dp.left = rDp.left;
			dp.left = (rMx.left + rVm.left) * inf->ph_ppi.x / inf->vm_ppi.x;
//			dp.top = dp_ry * dsr.x / dsr.y;
#ifdef USE_PAGE_BUFFER
			dp.top = (int)(vm_y * inf->ph_ppi.y * dsr.x / inf->vm_ppi.y / dsr.y);
#else
			dp.top = (int)(vm_ry * inf->ph_ppi.y * dsr.x / inf->vm_ppi.y / dsr.y);
#endif
			dp.right = (rMx.right + rVm.left) * inf->ph_ppi.x / inf->vm_ppi.x;
			dp.bottom = rMx.bottom * inf->ph_ppi.y * dsr.x / inf->vm_ppi.y / dsr.y;
#ifdef USE_TRIMING_AREA
			// 左端がはみ出る場合のトリミング
			if (dp.left < 1) {
				dp.left = 1;
				rMx.left = dp.left * inf->vm_ppi.x / inf->ph_ppi.x - rVm.left;
			}
			// 右端がはみ出る場合のトリミング
			if (dp.right + 2 > inf->ph_size.width) {
				dp.right = inf->ph_size.width - 2;
				rMx.right = dp.right * inf->vm_ppi.x / inf->ph_ppi.x - rVm.left;
			}
#endif
		} else if (inf->tech == capsPrinterType
			|| inf->tech == capsPreviewType) {
			// 印刷・印刷プレビュー
//			dp.left = rDp.left;
			dp.left = (rMx.left + rVm.left) * inf->ph_ppi.x / inf->vm_ppi.x;
//			dp.top = dp_ry;
			dp.top = (int)(vm_ry * inf->ph_ppi.y / inf->vm_ppi.y);
			dp.right = (rMx.right + rVm.left) * inf->ph_ppi.x / inf->vm_ppi.x;
			dp.bottom = rMx.bottom * inf->ph_ppi.y / inf->vm_ppi.y;
		}
#if 0
		wxString dbg;
		dbg = _T("put_buffer\n")
		+ wxString::Format(_T("rMx left:%d top:%d right:%d bottom:%d\n")
			,rMx.left, rMx.top, rMx.right, rMx.bottom)
		+ wxString::Format(_T("dp  left:%d top:%d right:%d bottom:%d\n")
			,dp.left, dp.top, dp.right, dp.bottom)
		;
		AddDebugLog(dbg);
#endif
		// 文字の入っている横幅と改行する高さ分だけの矩形を出力
		if (rMx.right > rMx.left && rMx.bottom > rMx.top) {
#ifdef USE_PIXELDATA
			dcMx->DrawBitmap(*bmpMx, 0, 0);
#else
			dcMx->DrawBitmap(
				wxBitmap(wxImage(MP_PRN_BUF_WIDTH,MP_PRN_BUF_HEIGHT,bufMx,true), MX_RGB_SIZE * 8)
				,0 ,0);
#endif
			dc->StretchBlit(dp.left, dp.top, dp.right - dp.left, dp.bottom
				,dcMx, rMx.left, rMx.top, rMx.right - rMx.left, rMx.bottom - rMx.top, inf->mBlitMode);
		}
	}
}


// 画面表示用 ページ番号と枠の描画
void MP_PRINTER::print_page_no(wxDC *dc, CapsInfoType *inf, int page)
{
	// ページ番号印刷用バッファ
	wxString str_page;
	wxPoint pos_page;
	size_t len_page;
	wxPoint wk1,wk2;
	wxSize fontsize = dc->GetFont().GetPixelSize();

	if (page < 1) return;

	// 枠を描画
	dc->SetPen(wxPen(wxColour(0, 0, 0), 1, wxPENSTYLE_SOLID));
	dc->SetBrush(wxBrush(wxColour(0xff, 0xff, 0xff), wxBRUSHSTYLE_SOLID));
#ifdef USE_PAGE_BUFFER
	wk1.x = 0;
	wk1.y = 0;
	wk1.y = wk1.y * dsr.x / dsr.y;
	wk2.x = inf->ph_size.width;
	wk2.y = inf->ph_size.height;
	wk2.y = wk2.y * dsr.x / dsr.y - wk1.y;
#else
	wk1.x = win_offset.x;
	wk1.y = inf->ph_size.height * (page - 1);
	wk1.y = wk1.y * dsr.x / dsr.y;
	wk2.x = inf->ph_size.width;
	wk2.y = inf->ph_size.height * page;
	wk2.y = wk2.y * dsr.x / dsr.y - wk1.y;
#endif
	dc->DrawRectangle(wk1.x, wk1.y, wk2.x, wk2.y);

	// 余白位置を描画
	dc->SetPen(wxPen(wxColour(0xc0,0xc0,0xc0),1,wxPENSTYLE_LONG_DASH));
#ifdef USE_PAGE_BUFFER
	wk1.x = 1;
	wk1.y = inf->ph_margin.top;
	wk1.y = wk1.y * dsr.x / dsr.y;
	wk2.x = inf->ph_size.width - 1;
#else
	wk1.x = win_offset.x + 1;
	wk1.y = inf->ph_size.height * (page - 1) + inf->ph_margin.top;
	wk1.y = wk1.y * dsr.x / dsr.y;
	wk2.x = inf->ph_size.width + win_offset.x - 1;
#endif
	if (inf->ph_margin.top != 0) {
		dc->DrawLine(wk1.x, wk1.y, wk2.x, wk1.y);
	}

#ifdef USE_PAGE_BUFFER
	wk1.x = 1;
	wk1.y = inf->ph_size.height - inf->ph_margin.bottom;
	wk1.y = wk1.y * dsr.x / dsr.y;
	wk2.x = inf->ph_size.width - 1;
#else
	wk1.x = win_offset.x + 1;
	wk1.y = inf->ph_size.height * page - inf->ph_margin.bottom;
	wk1.y = wk1.y * dsr.x / dsr.y;
	wk2.x = inf->ph_size.width + win_offset.x - 1;
#endif
	if (inf->ph_margin.bottom != 0) {
		dc->DrawLine(wk1.x, wk1.y, wk2.x, wk1.y);
	}

	// ページ番号印刷
	dc->SetTextForeground(wxColour(0xc0, 0xc0, 0xc0));
	str_page = _("Page ") + wxString::Format(_T("%d"), page);
	len_page = str_page.length();

#ifdef USE_PAGE_BUFFER
	pos_page.x = 2;
	pos_page.y = 2;
	pos_page.y = pos_page.y * dsr.x / dsr.y;
#else
	pos_page.x = 2 + win_offset.x;
	pos_page.y = 2 + (inf->ph_size.height * (page - 1));
	pos_page.y = pos_page.y * dsr.x / dsr.y;
#endif
	dc->DrawText(str_page, pos_page);

#ifdef USE_PAGE_BUFFER
	pos_page.x = inf->ph_size.width - (fontsize.GetX() * (len_page + 1)) - 2;
	pos_page.y = inf->ph_size.height - 2;
	pos_page.y = pos_page.y * dsr.x / dsr.y - fontsize.GetY();
#else
	pos_page.x = inf->ph_size.width - (fontsize.GetX() * (len_page + 1)) - 2 + win_offset.x;
	pos_page.y = inf->ph_size.height - 2 + (inf->ph_size.height * (page - 1));
	pos_page.y = pos_page.y * dsr.x / dsr.y - fontsize.GetY();
#endif
	dc->DrawText(str_page, pos_page);

}

// 印刷範囲の下端を計算
long MP_PRINTER::calc_dp_bottom(CapsInfoType *inf, int h)
{
	long bottom;

	if (inf->ph_size.height > 0) {
		bottom = inf->ph_size.height - inf->ph_offset.y - inf->ph_margin.bottom - (h * inf->ph_ppi.y / inf->vm_ppi.y);
		if (bottom > inf->ph_size.height) bottom = inf->ph_size.height;
	} else {
		bottom = -1;
	}
	return bottom;
}
long MP_PRINTER::calc_vm_bottom(CapsInfoType *inf, int h)
{
	long bottom;

	if (inf->vm_size.height > 0) {
		bottom = inf->vm_size.height - inf->vm_offset.y - inf->vm_margin.bottom - h;
		if (bottom > inf->vm_size.height) bottom = inf->vm_size.height;
	} else {
		bottom = -1;
	}
	return bottom;
}

// 画面表示用システムフォントの設定
void MP_PRINTER::set_sys_font(wxDC *dc, int w, int h)
{
	wxFont font(wxSize(w,h), wxFONTFAMILY_DEFAULT, wxFONTSTYLE_NORMAL, wxFONTWEIGHT_NORMAL, false, _T("Courier New"));
	dc->SetFont(font);
}

// ピクセルデータの内容を左へシフト
void MP_PRINTER::shift_pixeldata(int width)
{
	DpType *src = bufDw;
	DpType *dst = bufDw;

	if (width > MP_PRN_BUF_WIDTH) width = MP_PRN_BUF_WIDTH;
	if (width < 0) width = 0;

	for(int h=0; h < MP_PRN_BUF_HEIGHT; h++) {
		src = bufDw + h * MP_PRN_BUF_WIDTH + width;
		dst = bufDw + h * MP_PRN_BUF_WIDTH;
		for(int w=0; w < (MP_PRN_BUF_WIDTH - width); w++) {
			*dst = *src;
			src++;
			dst++;
		}
		for(int w=0; w < width; w++) {
			*dst = DpTypeMax;
			dst++;
		}
	}
}

// ピクセルデータの内容をクリア
void MP_PRINTER::clear_pixeldata(int width)
{
	DpType *src = bufDw;

	if (width > MP_PRN_BUF_WIDTH) width = MP_PRN_BUF_WIDTH;

	for(int h=0; h < MP_PRN_BUF_HEIGHT; h++) {
		src = bufDw + h * MP_PRN_BUF_WIDTH;
		for(int w=0; w < width; w++) {
			*src = DpTypeMax;
			src++;
		}
	}
}

// ピクセルデータの内容をコピー
void MP_PRINTER::copy_pixeldata(int dx, int dy, int width, int height, int sx, int sy)
{
	DpType *src = bufDw;
#ifdef USE_PIXELDATA
	wxNativePixelData::Iterator dst(*bufMx);
#else
	wxByte *dst = bufMx;
#endif

	if (width + wxMax(sx,dx) > MP_PRN_BUF_WIDTH) width = MP_PRN_BUF_WIDTH - wxMax(sx,dx);
	if (height + wxMax(sy,dy) > MP_PRN_BUF_HEIGHT) height = MP_PRN_BUF_HEIGHT - wxMax(sy,dy);

	for(int h=0; h < height; h++) {
		src = bufDw + (sy + h) * MP_PRN_BUF_WIDTH + sx;
#ifdef USE_PIXELDATA
		dst.MoveTo(*bufMx, dx, dy + h);
#else
		dst = bufMx + ((dy + h) * MP_PRN_BUF_WIDTH + dx) * MX_RGB_SIZE;
#endif
		for(int w=0; w < width; w++) {
#ifdef USE_PIXELDATA
			dst.Red() = (*src) & dst.Red();
			dst.Green() = (*src) & dst.Green();
			dst.Blue() = (*src) & dst.Blue();
			src++;
			dst++;
#else
			for(int n=0; n<MX_RGB_SIZE; n++) {
				dst[n] = (*src) & dst[n];
			}
			src++;
			dst+=MX_RGB_SIZE;
#endif
		}
	}
}

// 合成エリアをを上へシフト
void MP_PRINTER::shift_mixbuffer(int width, int height)
{
#ifdef USE_PIXELDATA
	wxNativePixelData::Iterator src(*bufMx);
	wxNativePixelData::Iterator dst(*bufMx);
#else
	wxByte *src = bufMx;
	wxByte *dst = bufMx;
#endif
	if (height > MP_PRN_BUF_HEIGHT) height = MP_PRN_BUF_HEIGHT;
	if (height < 0) height = 0;

	for(int h=0; h < (MP_PRN_BUF_HEIGHT - height); h++) {
#ifdef USE_PIXELDATA
		src.MoveTo(*bufMx,0,h + height);
		dst.MoveTo(*bufMx,0,h);
#else
		src = bufMx + ((h + height) * MP_PRN_BUF_WIDTH) * MX_RGB_SIZE;
		dst = bufMx + (h * MP_PRN_BUF_WIDTH) * MX_RGB_SIZE;
#endif
		for(int w=0; w < width; w++) {
#ifdef USE_PIXELDATA
			dst.Red() = src.Red();
			dst.Green() = src.Green();
			dst.Blue() = src.Blue();
			src++;
			dst++;
#else
			for(int n=0; n<MX_RGB_SIZE; n++) {
				dst[n] = src[n];
			}
			src+=MX_RGB_SIZE;
			dst+=MX_RGB_SIZE;
#endif
		}
		rMxMax[h].left = rMxMax[h + height].left;
		rMxMax[h].right = rMxMax[h + height].right;
	}
	for(int h=(MP_PRN_BUF_HEIGHT - height); h < MP_PRN_BUF_HEIGHT; h++) {
#ifdef USE_PIXELDATA
		dst.MoveTo(*bufMx,0,h);
#else
		dst = bufMx + (h * MP_PRN_BUF_WIDTH) * MX_RGB_SIZE;
#endif
		for(int w=0; w < width; w++) {
#ifdef USE_PIXELDATA
			dst.Red() = dst.Green() = dst.Blue() = DpTypeMax;
			dst++;
#else
			for(int n=0; n<MX_RGB_SIZE; n++) {
				dst[n] = DpTypeMax;
			}
			dst+=MX_RGB_SIZE;
#endif
		}
		rMxMax[h].left = width;
		rMxMax[h].right = 0;
	}
}

// 合成エリアををクリア
void MP_PRINTER::clear_mixbuffer(int width)
{
#ifdef USE_PIXELDATA
	wxNativePixelData::Iterator dst(*bufMx);
#else
	wxByte *dst = bufMx;
#endif
	for(int h=0; h < MP_PRN_BUF_HEIGHT; h++) {
#ifdef USE_PIXELDATA
		dst.MoveTo(*bufMx, 0, h);
#else
		dst = bufMx + (h * MP_PRN_BUF_WIDTH) * MX_RGB_SIZE;
#endif
		for(int w=0; w < width; w++) {
#ifdef USE_PIXELDATA
			dst.Red() = dst.Green() = dst.Blue() = DpTypeMax;
			dst++;
#else
			for(int n=0; n<MX_RGB_SIZE; n++) {
				dst[n] = DpTypeMax;
			}
			dst+=MX_RGB_SIZE;
#endif
		}
		rMxMax[h].left = width;
		rMxMax[h].right = 0;
	}
}

#ifdef USE_CASH_BUFFER
// 保存しているデータで描画
bool MP_PRINTER::draw_from_cashdata(wxDC *dc, CapsInfoType *inf)
{
	int win_offset_x = (win_size.width - inf->ph_size.width) / 2;
	if (win_offset_x < 0) win_offset_x = 0;
	dc->Blit(win_offset_x,0,bmpCashd->GetWidth(),bmpCashd->GetHeight(),dcCashd,0,0);
	return true;
}

// 保存用デバイスに描画
bool MP_PRINTER::draw_to_cashdata(wxDC *dc, CapsInfoType *inf)
{
	draw_data(dcCashd, inf);
	return draw_from_cashdata(dc, inf);
}
#endif

//
// 印刷制御クラス
//
MpPrintout::MpPrintout(MP_PRINTER *new_mpri, const wxString &title, wxPageSetupDialogData *new_pagesetdlg)
	: wxPrintout(title)
{
	mpri = new_mpri;
	pagesetdlg = new_pagesetdlg;
}

// 印刷処理開始
void MpPrintout::OnBeginPrinting()
{
	AddDebugLog(_T("OnBeginPrinting"));

	pdc = this->GetDC();

	mpri->PrintingStart(pdc, IsPreview());

	wxPoint scr_ppi;
	wxPoint prn_ppi;
	GetPPIScreen(&scr_ppi.x, &scr_ppi.y);
	GetPPIPrinter(&prn_ppi.x, &prn_ppi.y);
	wxRect page_rect = GetLogicalPageRect();
	wxRect paper_rect = GetLogicalPaperRect();
	wxRect r_paper_rect = GetPaperRectPixels();
	MySize px;
	GetPageSizePixels(&px.width, &px.height);
	wxRect r_page_rect(0,0, px.width, px.height);

#ifdef _DEBUG_LOG
	wxString dbg;
	wxString cname = pdc->GetClassInfo()->GetClassName();
	dbg = _T("DC: ") + cname + _T("\n");
	dbg = dbg + _T("Before\n");
	dbg = dbg + wxString::Format(_T("  screen ppi x:%d y:%d\n"),scr_ppi.x,scr_ppi.y);
	dbg = dbg + wxString::Format(_T("  printer ppi x:%d y:%d\n"),prn_ppi.x,prn_ppi.y);
	dbg = dbg + wxString::Format(_T("  printer page size  w:%d h:%d\n"),r_page_rect.GetWidth(),r_page_rect.GetHeight());
	dbg = dbg + wxString::Format(_T("  printer paper rect left:%d top:%d width:%d height:%d\n"),r_paper_rect.GetLeft(),r_paper_rect.GetTop(),r_paper_rect.GetWidth(),r_paper_rect.GetHeight());
	dbg = dbg + wxString::Format(_T("  printer loc page size width:%d height:%d\n"),page_rect.GetWidth(),page_rect.GetHeight());
	dbg = dbg + wxString::Format(_T("  printer loc paper rect left:%d top:%d width:%d height:%d\n"),paper_rect.GetLeft(),paper_rect.GetTop(),paper_rect.GetWidth(),paper_rect.GetHeight());
	AddDebugLog(dbg);
#endif

	if (IsPreview()) {
		// 印刷プレビュー時
#if defined(__WXMSW__)
		mpri->SetDocumentSize(&scr_ppi, &prn_ppi, &paper_rect, &page_rect);
#elif defined(__WXOSX__)
		MapScreenSizeToPaper();
		MapScreenSizeToPage();

		GetPPIScreen(&scr_ppi.x, &scr_ppi.y);
		GetPPIPrinter(&prn_ppi.x, &prn_ppi.y);
		page_rect = GetLogicalPageRect();
		paper_rect = GetLogicalPaperRect();
		r_paper_rect = GetPaperRectPixels();
		GetPageSizePixels(&px.width, &px.height);
		r_page_rect = wxRect(0,0, px.width, px.height);

		mpri->SetDocumentSize(&scr_ppi, &scr_ppi, &paper_rect, &page_rect);
#elif defined(__WXGTK__)
		MapScreenSizeToPaper();
		MapScreenSizeToPage();

		GetPPIScreen(&scr_ppi.x, &scr_ppi.y);
		GetPPIPrinter(&prn_ppi.x, &prn_ppi.y);
		page_rect = GetLogicalPageRect();
		paper_rect = GetLogicalPaperRect();
		r_paper_rect = GetPaperRectPixels();
		GetPageSizePixels(&px.width, &px.height);
		r_page_rect = wxRect(0,0, px.width, px.height);

		mpri->SetDocumentSize(&scr_ppi, &scr_ppi, &paper_rect, &page_rect);
#endif
	} else {
		// 印刷時
#if defined(__WXMSW__)
		mpri->SetDocumentSize(&scr_ppi, &prn_ppi, &paper_rect, &page_rect);
#elif defined(__WXOSX__)
		mpri->SetDocumentSize(&scr_ppi, &prn_ppi, &paper_rect, &page_rect);
#elif defined(__WXGTK__)
		MapScreenSizeToPaper();
		MapScreenSizeToPage();

		GetPPIScreen(&scr_ppi.x, &scr_ppi.y);
		GetPPIPrinter(&prn_ppi.x, &prn_ppi.y);
		page_rect = GetLogicalPageRect();
		paper_rect = GetLogicalPaperRect();
		r_paper_rect = GetPaperRectPixels();
		GetPageSizePixels(&px.width, &px.height);
		r_page_rect = wxRect(0,0, px.width, px.height);

		// TODO: StretchBlitでスケーリングがおかしい／なぜ？
		// screen: 72ppiに換算して計算する。
		prn_ppi.x = prn_ppi.x * scr_ppi.x / 72;
		prn_ppi.y = prn_ppi.y * scr_ppi.y / 72;
		r_paper_rect.x = r_paper_rect.x * scr_ppi.x / 72;
		r_paper_rect.y = r_paper_rect.y * scr_ppi.y / 72;
		r_paper_rect.width = r_paper_rect.width * scr_ppi.x / 72;
		r_paper_rect.height = r_paper_rect.height * scr_ppi.y / 72;
		r_page_rect.x = r_page_rect.x * scr_ppi.x / 72;
		r_page_rect.y = r_page_rect.y * scr_ppi.y / 72;
		r_page_rect.width = r_page_rect.width * scr_ppi.x / 72;
		r_page_rect.height = r_page_rect.height * scr_ppi.y / 72;

		mpri->SetDocumentSize(&scr_ppi, &prn_ppi, &r_paper_rect, &r_page_rect);
#endif
	}

#ifdef _DEBUG_LOG
	dbg = _T("After\n");
	dbg = dbg + wxString::Format(_T("  screen ppi x:%d y:%d\n"),scr_ppi.x,scr_ppi.y);
	dbg = dbg + wxString::Format(_T("  printer ppi x:%d y:%d\n"),prn_ppi.x,prn_ppi.y);
	dbg = dbg + wxString::Format(_T("  printer page size  w:%d h:%d\n"),r_page_rect.GetWidth(),r_page_rect.GetHeight());
	dbg = dbg + wxString::Format(_T("  printer paper rect left:%d top:%d width:%d height:%d\n"),r_paper_rect.GetLeft(),r_paper_rect.GetTop(),r_paper_rect.GetWidth(),r_paper_rect.GetHeight());
	dbg = dbg + wxString::Format(_T("  printer loc page size width:%d height:%d\n"),page_rect.GetWidth(),page_rect.GetHeight());
	dbg = dbg + wxString::Format(_T("  printer loc paper rect left:%d top:%d width:%d height:%d\n"),paper_rect.GetLeft(),paper_rect.GetTop(),paper_rect.GetWidth(),paper_rect.GetHeight());
	AddDebugLog(dbg);
#endif

}
// ドキュメント印刷開始
bool MpPrintout::OnBeginDocument(int startPage, int endPage)
{
	AddDebugLog(_T("OnBeginDocument"));
//	if (!wxPrintout::OnBeginDocument(startPage, endPage))
//		return false;

	return mpri->PrintStartDocument(pdc);
}
// 印刷中
bool MpPrintout::OnPrintPage(int pageNum)
{
	AddDebugLog(wxString::Format(_T("OnPrintPage %d"), pageNum));
	WritePageHeader(this, pdc, wxEmptyString, 1.0);
	return mpri->PrintPage(pdc, pageNum);
}
// ドキュメント印刷終了
void MpPrintout::OnEndDocument()
{
	AddDebugLog(_T("OnEndDocument"));
	mpri->PrintEndDocument(pdc);
}
// 印刷処理終了
void MpPrintout::OnEndPrinting()
{
	AddDebugLog(_T("OnEndPrinting"));
	mpri->PrintingEnd(pdc);
}
// 印刷ページ設定
void MpPrintout::GetPageInfo(int *minPage, int *maxPage, int *selPageFrom, int *selPageTo)
{
	mpri->GetPageInfo(minPage, maxPage, selPageFrom, selPageTo);
	AddDebugLog(wxString::Format(_T("GetPageInfo: minpage:%d maxpage:%d frompage:%d topage:%d"),*minPage,*maxPage,*selPageFrom,*selPageTo));
}
// 印字対象ページ？
bool MpPrintout::HasPage(int pageNum)
{
    return true;
}
// ヘッダ
bool MpPrintout::WritePageHeader(wxPrintout *printout, wxDC *dc, const wxString&text, float mmToLogical)
{
	wxRect prect = GetLogicalPaperRect();
	dc->SetPen(wxPen(wxColour(0xff,0xff,0xff), 1));
	dc->SetBrush(wxBrush(wxColour(0xff, 0xff, 0xff), wxBRUSHSTYLE_SOLID));
	dc->DrawRectangle(prect.GetLeft(),prect.GetTop(),prect.GetWidth(),prect.GetHeight());
	prect = GetLogicalPageRect();
	dc->SetPen(wxPen(wxColour(0xc0,0xc0,0xc0), 1, wxPENSTYLE_DOT));
	if (IsPreview()) {
		dc->DrawRectangle(prect.GetLeft(),prect.GetTop(),prect.GetWidth(),prect.GetHeight());
	}
#ifdef _DEBUG
	dc->DrawRectangle(prect.GetLeft(),prect.GetTop(),prect.GetWidth(),prect.GetHeight());
	dc->DrawLine(prect.GetLeft(),prect.GetTop(),prect.GetRight(),prect.GetBottom());
	dc->DrawLine(prect.GetLeft(),prect.GetBottom(),prect.GetRight(),prect.GetTop());
#endif

	return true;
}

//
// 印刷プレビュークラス
//
MpPreviewFrame::MpPreviewFrame(wxPrintPreview *preview, wxWindow *parent)
	: wxPreviewFrame(preview, parent, _("Preview"), wxPoint(100, 100), wxSize(600, 650))
{
	frame = parent;
	Centre(wxBOTH);
	InitializeWithModality(wxPreviewFrame_WindowModal);
	Show();
}
MpPreviewFrame::~MpPreviewFrame()
{
	frame->SetFocus();
}

